package org.jaggregator.generator;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Collection;
import java.util.Date;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.collections.ExtendedProperties;
import org.apache.commons.lang.exception.ExceptionUtils;
import org.apache.log4j.Category;
import org.apache.velocity.VelocityContext;
import org.apache.velocity.app.Velocity;
import org.jaggregator.Main;
import org.jaggregator.config.SubscriptionListDecl;
import org.jaggregator.config.TemplateDecl;
import org.jaggregator.util.ItemUtil;

/**
 * Generates the site.
 *
 * @author <a href="mailto:jsz@jaggregator.org">Janek Schwarz</a>
 * @version $Revision: 1.13 $
 */
public class SiteGenerator 
{
private static final String LOG4J_VELOCITY_CATEGORY =
  "org.jaggregator.Velocity";

private static final String VELOCITY_PROPERTIES_RESOURCE =
  "resources/org/jaggregator/velocity.properties";

private Category log = Category.getInstance(SiteGenerator.class);
private Map allSubscriptionsMap;
private File skinDirectory;
private File outputDirectory;
private Collection subscriptionListCol;
private Collection templateDeclCol;

public SiteGenerator(Map allSubscriptionsMap,
                     File skinDirectory,
                     File outputDirectory,
                     Collection subscriptionListCol,
                     Collection templateDeclCol) 
{
  this.allSubscriptionsMap = allSubscriptionsMap;
  this.skinDirectory = skinDirectory;
  this.outputDirectory = outputDirectory;
  this.subscriptionListCol = subscriptionListCol;
  this.templateDeclCol = templateDeclCol;
}

public void initialize()
  throws SiteGenerationException
{
  initializeVelocity();
  sanityCheckSubscriptionListDecls();
  sanityCheckTemplateDecls();
}

private void initializeVelocity()
  throws SiteGenerationException
{
  Category velocityLog = Category.getInstance(LOG4J_VELOCITY_CATEGORY);
  
  ClassLoader threadClassLoader = 
    Thread.currentThread().getContextClassLoader();
  InputStream velocityConfigurationIstream = 
    threadClassLoader.getResourceAsStream(VELOCITY_PROPERTIES_RESOURCE);

  if(velocityConfigurationIstream == null)
    {
      throw new SiteGenerationException(
        "Can't find " + VELOCITY_PROPERTIES_RESOURCE);
    }

  try
    {
      ExtendedProperties velocityProperties = new ExtendedProperties();
      velocityProperties.load(velocityConfigurationIstream);
      velocityProperties.addProperty("file.resource.loader.path", 
                                     skinDirectory.getPath());
      Velocity.setExtendedProperties(velocityProperties);
      Velocity.init();
    }
  catch(IOException ex)
    {
      throw new SiteGenerationException(
        "Can't read " + VELOCITY_PROPERTIES_RESOURCE,
        ex);
    }
  catch(Exception ex)
    {
      throw new SiteGenerationException(
        "General exception during setup of velocity", ex);
    }
}

private void sanityCheckSubscriptionListDecls()
  throws SiteGenerationException
{
  if(subscriptionListCol.size() == 0)
    throw new SiteGenerationException(
      "Configuration does not contain any subscription list definition.");
}

private void sanityCheckTemplateDecls()
  throws SiteGenerationException
{
  if(templateDeclCol.size() == 0)
    throw new SiteGenerationException(
      "Configuration does not contain any template definition.");

  Iterator templateDeclsIter = templateDeclCol.iterator();
  while(templateDeclsIter.hasNext())
    {
      TemplateDecl decl = (TemplateDecl) templateDeclsIter.next();
      Collection refs = decl.getSubscriptionListRefs();
      
      if(refs.size() == 0)
        {
          log.warn("The template declaration '" + decl.getFileName() +
                   "' does not contain any references to subscription lists");

          continue;
        }
      
      Iterator refsIter = refs.iterator();
      while(refsIter.hasNext())
        {
          String ref = (String) refsIter.next();

          Object subscriptionListDecl = 
            CollectionUtils.find(
              subscriptionListCol,
              new SubscriptionListDeclByReferencePredicate(ref));
          
          if(subscriptionListDecl == null)
            {
              log.warn("The template '" + decl.getFileName() +
                       "' contains a reference to subscription list named '" +
                       ref + "' that is not defined");
            }
        }

      String clazzName = decl.getInvokeModeClassName();
      try
        {
          Class clazz = Class.forName(clazzName);
          if(Invoker.class.isAssignableFrom(clazz) == false)
            {
              log.error("invokeModeClass '" + clazzName +
                        "' does not implement the interface Invoker.");
            }

          Object invoker = clazz.newInstance();
        }
      catch(ClassNotFoundException ex)
        {
          log.error("invokeModeClass '" + clazzName +
                    "' cannot be found.");
          log.debug(ExceptionUtils.getStackTrace(ex));
        }
      catch(InstantiationException ex)
        {
          log.error("invokeModeClass '" + clazzName +
                    "' cannot be instantiated.");
          log.debug(ExceptionUtils.getStackTrace(ex));
        }
      catch(IllegalAccessException ex)
        {
          log.error("invokeModeClass '" + clazzName +
                    "' cannot be instantiated.");
          log.debug(ExceptionUtils.getStackTrace(ex));
        }
    }
}

public void generate(Collection updatedChannelUris)
  throws SiteGenerationException
{
  log.debug("generate site");
  log.debug("updated channels: " + updatedChannelUris);
  
  Collection subscriptionListsContainingUris = 
    selectSubscriptionListsContainingUris(updatedChannelUris);
  log.debug("subscription list containing updated channels: " + 
            subscriptionListsContainingUris);
  
  Collection templateCol =
    selectTemplateDeclsReferencing(subscriptionListsContainingUris);
  log.debug("templates referencing subscription lists: " +
            templateCol);

  Date now = new Date();
  Iterator templateIter = templateCol.iterator();
  while(templateIter.hasNext())
    {
      TemplateDecl decl = (TemplateDecl) templateIter.next();
      log.debug("processing template decl: " + decl.getFileName());
      
      Collection referencedSubscriptions =
        selectSubscriptionsReferencedBy(decl);
      log.debug("template decl references subscriptions: " + 
                referencedSubscriptions);

      String clazzName = decl.getInvokeModeClassName();
      try
        {
          VelocityContext context = new VelocityContext();
          context.put("jaggregatorVersion", Main.VERSION_STRING);
          context.put("skinDirectoryUrl", skinDirectory.toURL());
          context.put("now", now);
          context.put("itemUtil", new ItemUtil());
          
          Class clazz = Class.forName(clazzName);
          Invoker invoker = (Invoker) clazz.newInstance();

          invoker.setOutputDirectory(outputDirectory);
          invoker.apply(
            decl, 
            referencedSubscriptions,
            subscriptionListCol,
            allSubscriptionsMap,
            context);
        }
      catch(ClassNotFoundException ex)
        {
          log.error("invokeModeClass '" + clazzName +
                    "' cannot be found.");
          log.debug(ExceptionUtils.getStackTrace(ex));
        }
      catch(InstantiationException ex)
        {
          log.error("invokeModeClass '" + clazzName +
                    "' cannot be instantiated.");
          log.debug(ExceptionUtils.getStackTrace(ex));
        }
      catch(IllegalAccessException ex)
        {
          log.error("invokeModeClass '" + clazzName +
                    "' cannot be instantiated.");
          log.debug(ExceptionUtils.getStackTrace(ex));
        }
      catch(MalformedURLException ex)
        {
          log.error("can't convert skinDirectory '" + 
                    skinDirectory.getAbsolutePath() +
                    "' to an URL");
          log.debug(ExceptionUtils.getStackTrace(ex));
        }
    }

  log.debug("site generated");
}

private Collection selectSubscriptionListsContainingUris(Collection uriCol)
{
  Iterator iter = uriCol.iterator();
  Set subscriptionListDeclSet = new HashSet();
  while(iter.hasNext())
    {
      URL channelUri = (URL) iter.next();

      Collection col = CollectionUtils.select(
        subscriptionListCol,
        new SubscriptionListDeclByChannelUriPredicate(channelUri));

      if(col.size() == 0)
        {
          log.warn(
            "Can't find a subscription list containing a subsription to '" +
            channelUri.toExternalForm()
            + "'. This could be programming error");
        }
      else
        {
          subscriptionListDeclSet.addAll(col);
        }
    }

  return subscriptionListDeclSet;
}

private Collection selectTemplateDeclsReferencing(Collection subscriptionsCol)
{
  Iterator iter = subscriptionsCol.iterator();
  Set templateDeclSet = new HashSet();
  while(iter.hasNext())
    {
      SubscriptionListDecl decl = (SubscriptionListDecl) iter.next();
      
      Collection col = 
        CollectionUtils.select(
          templateDeclCol,
          new TemplateDeclBySubscriptionListDeclRefPredicate(decl.getId()));

      if(col.size() == 0)
        {
          log.warn(
            "Can't find a template referencing the subscription list '" +
            decl.getId() +
            "'. This could be programming error");
        }
      else
        {
          templateDeclSet.addAll(col);
        }
      
    }

  return templateDeclSet;
}

private Collection selectSubscriptionsReferencedBy(TemplateDecl decl)
{
  Collection referencedSubscriptionsCol = new HashSet();
  
  Iterator iter = decl.getSubscriptionListRefs().iterator();
  while(iter.hasNext())
    {
      String ref = (String) iter.next();

      SubscriptionListDecl subscriptionListDecl = 
        (SubscriptionListDecl) CollectionUtils.find(
          subscriptionListCol,
          new SubscriptionListDeclByReferencePredicate(ref));
      
      if(subscriptionListDecl != null)
        {
          Set channelUris = subscriptionListDecl.keySet();

          Iterator uriIter = channelUris.iterator();
          while(uriIter.hasNext())
            {
              URL channelUri = (URL) uriIter.next();
              
              if(allSubscriptionsMap.containsKey(channelUri))
                {
                  referencedSubscriptionsCol.add(
                    allSubscriptionsMap.get(channelUri));
                }
            }
        }
    }

  return referencedSubscriptionsCol;
}
}
