package org.jaggregator.generator;

import java.text.DateFormat;
import java.util.Calendar;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.TreeSet;
import org.apache.commons.collections.MultiHashMap;
import org.apache.log4j.Category;
import org.apache.velocity.VelocityContext;
import org.jaggregator.rss.Item;
import org.jaggregator.rss.Subscription;
import org.jaggregator.util.DateUtils;

/**
 * The <code>TodayView</code> builds a structure that contains
 * only those items of channels that have been retrieved during
 * the last 24 hours. In particular, the structure maps
 * <code>Date</code> objects to a list of <code>Item</code>
 * instances that have been retrieved at that date.
 *
 * @author <a href="mailto:jsz@jaggregator.de">Janek Schwarz</a>
 * @version $Revision: 1.6 $
 */
public class TodayView 
  implements AllListsView
{
private static final String DATE_FORMAT_PATTERN = "HH:mm:ss";
private Category log = Category.getInstance(TodayView.class);
private VelocityContext context;
private Collection subscriptionsCol;

public TodayView() 
{}

public void setContext(VelocityContext context)
{
  this.context = context;
}

public void setSubscriptions(Collection subscriptionsCol)
{
  this.subscriptionsCol = subscriptionsCol;
}

public void initialize()
{
  Calendar calendar = Calendar.getInstance();
  calendar.add(Calendar.DAY_OF_MONTH, -1);
  Date yesterday = DateUtils.getEvenMinuteDateBefore(calendar.getTime());

  MultiHashMap dateItemMap = 
    new MultiHashMap()
    {
    public Set keySet()
    {
      Set set = new TreeSet(Collections.reverseOrder());
      set.addAll(super.keySet());
      
      return set;
    }
    };
  
  Iterator subscriptionsIter = subscriptionsCol.iterator();
  while(subscriptionsIter.hasNext())
    {
      Subscription subscription = (Subscription) subscriptionsIter.next();
  
      Set changedDatesSet = new TreeSet(Collections.reverseOrder());
      changedDatesSet.addAll(subscription.getChangedDatesSince(yesterday));
      
      Iterator changedDatesIter = changedDatesSet.iterator();
      Date beforeDate = null;
      log.debug("subscription: " + subscription.getUrl());
      while(changedDatesIter.hasNext())
        {
          Date changedDate = (Date) changedDatesIter.next();
          List items;

          log.debug("-----------");
          log.debug("beforeDate: " + beforeDate);
          log.debug("changedDate: " + changedDate);

          if(beforeDate == null)
            items = subscription.getItems(changedDate);
          else
            items = subscription.getItems(changedDate, beforeDate);
          
          beforeDate = changedDate;
          
          Date evenMinuteDate = DateUtils.getEvenMinuteDateBefore(changedDate);
          dateItemMap.put(evenMinuteDate,
                          new Tuple(subscription, items));
        }
    }

  context.put("todayDateItemMap", dateItemMap);
  context.put("todayDateFormat", 
              DateFormat.getTimeInstance(DateFormat.MEDIUM));
  
  context.put("isToday", new IsToday());
}

public class Tuple
{
private Subscription subscription;
private List itemList;

public Tuple(Subscription subscription, List itemList)
{
  this.subscription = subscription;
  this.itemList = itemList;
}

public Subscription getSubscription()
{
  return subscription;
}

public List getItemList()
{
  return itemList;
}
}

public class IsToday
{
public boolean isToday(Date date)
{
  Date today = DateUtils.getDateOf(0, 0, 0);
  
  if(date.compareTo(today) < 0)
    return false;
  else
    return true;
}
}
}
