package org.jaggregator.rss;

import java.net.URL;
import java.util.Collection;
import java.util.Date;
import java.util.List;
import java.util.Map;

/**
 * An instance of <code>Subscription</code> holds informations
 * about a channel, such as the channel's location (URL), the
 * date when the channel was retrieved the last time, if the last
 * retrieval succeded and retrieved new items etc. 
 *
 * @author <a href="mailto:jsz@jaggregator.org">Janek Schwarz</a>
 * @version $Revision: 1.12 $
 */
public interface Subscription
{
/**
 * Get the URL at which the channel is located.
 */
URL getUrl();

/**
 * Set the URL at which the channel is located.
 */
void setUrl(URL url);

/**
 * Get the name. If no specific name is set, use the title of the
 * channel.
 *
 * @return the name
 */
String getName();

/**
 * Set the name. It makes sense to set the name, if the channel's
 * name is inappropriate or too long or whatnot.
 *
 * @param name the name
 */
void setName(String name);

/**
 * Get the date when the channel was subscribed to.
 */
Date getSubscriptionDate();

/**
 * Set the date when the channel was subscribed to.
 */
void setSubscriptionDate(Date subscriptionDate);

/**
 * Get the channel.
 */
Channel getChannel();

/**
 * Set the channel.
 */
void setChannel(Channel channel);

/**
 * Update this instances channel with the informations contained
 * in the parameter. LastRetrievedDate is implicitely updated.
 */
void updateChannel(Channel channel);

/**
 * Get the date when the channel was retrieved the last time.
 */
Date getLastRetrievedDate();

/**
 * Set the date when the channel was retrieved the last time.
 */
void setLastRetrievedDate(Date lastRetrievedDate);

/**
 * Test whether the channel has changed since a certain date.
 */
boolean hasChangedSince(Date since);

/**
 * Return a collection of dates at which the channel has changed
 * since a given date. The order of the dates is not specified.
 *
 * @param since a <code>Date</code> value
 * @return a <code>Collection</code> value
 */
Collection getChangedDatesSince(Date since);

/**
 * Test whether the last retrieval of the channel retrieved new items.
 */
boolean hasNewItems();

/**
 * Get items which were retrieved after a certain date.
 */
List getItems(Date after);

/**
 * Get items which were retrieved after <code>afterDate</code>
 * but before <code>beforeDate</code>.
 */
List getItems(Date beforeDate, Date afterDate);

/**
 * Get the items which were retrieved when the channel was
 * retrieved the last time.
 */
List getNewItems();

/**
 * Get a collection that maps dates to a list of items that were
 * received at that date.
 *
 * @return a <code>Map</code> value
 */
Map getDateItemMap();

/**
 * Return true if the channel has changes that are not item level
 * changes. This includes for instance the channel title or
 * syndication information.
 */
boolean hasChannelLevelChanges();

/**
 * Set <code>lastModifiedString</code>.
 * <code>lastModifiedString</code> is the string that the HTTP
 * response header &quot;Last-Modified&quot; contained at the
 * time when the channel was retrieved.
 */
void setLastModifiedString(String lastModifiedString);

/**
 * Get <code>lastModifiedString</code>.
 * <code>lastModifiedString</code> is the string that the
 * HTTP response header &quot;Last-Modified&quot; contained at
 * the time when the channel was retrieved.
 */
String getLastModifiedString();

/**
 * Test whether lastModifiedString is set.
 */
boolean hasLastModifiedString();

/**
 * Set <code>eTagString</code>. <code>eTagString</code> is the
 * string that the HTTP response header &quot;ETag&quot;
 * contained at the time when the channel was retrieved.
 */
void setEtagString(String eTagString);

/**
 * Get <code>eTagString</code>. <code>eTagString</code> is the
 * string that the HTTP response header &quot;ETag&quot;
 * contained at the time when the channel was retrieved.
 */
String getEtagString();

/**
 * Test whether eTagString is set.
 */
boolean hasEtagString();
}
