package org.jaggregator.rss.common;

import java.util.Collection;
import java.util.Iterator;
import org.apache.commons.beanutils.ConvertUtils;
import org.apache.commons.digester.Digester;
import org.apache.commons.digester.RuleSet;
import org.jaggregator.rss.Converter;
import org.jaggregator.rss.ConverterSet;

/**
 * A <code>DigesterSupport</code> provides a couple of
 * convinience and template methodes for classes that parse XML
 * content using Jakarta's digester.
 *
 * @author <a href="mailto:jsz@jaggregator.org">Janek Schwarz</a>
 * @version $Revision: 1.2 $
 */
public abstract class DigesterSupport 
{
/**
 * The digester to use for parsing XML content
 */
protected Digester digester;

protected DigesterSupport()
{}

/**
 * Set up this instance. Subclasses have to call this method
 * before they start using the digester.
 */
protected void setup()
{
  setupConverters();
  setupDigester();
}

private void setupDigester()
{
  digester = new Digester();

  digester.setNamespaceAware(shouldSupportNamespaces());
    
  Iterator iter = getDigesterRuleSetCollection().iterator();
  while(iter.hasNext())
    digester.addRuleSet((RuleSet) iter.next());
}

private void setupConverters()
{
  ConverterSet converterSet = getConverterSet();
  
  Converter[] converters = converterSet.getConverters();
  for(int i = 0; i < converters.length; i++)
    {
      Converter converter = converters[i];
      if(converter != null)
        ConvertUtils.register(converter, converter.getClazz());
    }
}

/**
 * Subclasses could overide this method in order to provide
 * converters that the digester uses for converting content from
 * the XML instance to actual java objects.
 *
 * @return a <code>ConverterSet</code> value. May not be null. If
 * not overwritten, it returns an <code>EmptyConverterSet</code>.
 * @see BeanUtils.MethodUtils
 */
protected ConverterSet getConverterSet()
{
  return new EmptyConverterSet();
}

/**
 * Provide the digester with a set of rules.
 *
 * @return a <code>Collection</code> of
 * <code>RuleSet</code>s. May not be null.
 */
protected abstract Collection getDigesterRuleSetCollection();

/**
 * Subclasses may override this method in order to configure
 * namespace support of the digester. Per default, this methode
 * returns <code>false</code>.
 *
 * @return false.
 */
protected boolean shouldSupportNamespaces()
{
  return false;
}
}
