package org.jaggregator.rss.common;

import java.net.URL;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;
import org.apache.commons.collections.MultiHashMap;
import org.apache.commons.collections.MultiMap;
import org.apache.commons.lang.builder.ToStringBuilder;
import org.apache.log4j.Category;
import org.jaggregator.rss.Channel;
import org.jaggregator.rss.Item;
import org.jaggregator.rss.Subscription;

/**
 * Implementation of the <code>Subscription</code> interface.
 *
 * @author <a href="mailto:jsz@jaggregator.org">Janek Schwarz</a>
 * @version $Revision: 1.19 $
 */
public class SubscriptionImpl
  implements Subscription
{
private Category log = Category.getInstance(SubscriptionImpl.class);
private URL url;
private String name;
private Date subscriptionDate;
private Channel channel;
private Date lastRetrievedDate;
private boolean channelLevelChanges;
private String lastModifiedString;
private String eTagString;

MultiHashMap dateItemMap;

public SubscriptionImpl()
{
  dateItemMap = new MultiHashMap();
}

public synchronized URL getUrl()
{
  return url;
}

public synchronized void setUrl(URL url)
{
  this.url = url;
}

public synchronized String getName()
{
  String ret = name;
  
  if(name == null && channel != null)
    {
      ret = channel.getTitle();
    }
  
  if(ret == null)
    {
      ret = getUrl().toExternalForm();
    }
  
  return ret;
}

public synchronized void setName(String name)
{
  this.name = name;
}

public synchronized Date getSubscriptionDate()
{
  return subscriptionDate;
}

public synchronized void setSubscriptionDate(Date subscriptionDate)
{
  this.subscriptionDate = subscriptionDate;
}

public synchronized Channel getChannel()
{
  return channel;
}

public synchronized void setChannel(Channel channel)
{
  List channelItems = channel.getItems();
  Date lastRetrieved = getLastRetrievedDate();
  
  Iterator iter = channelItems.iterator();
  while(iter.hasNext())
    {
      Item item = (Item) iter.next();
      
      dateItemMap.put(lastRetrieved, item);
    }

  this.channel = channel;
}

void setChannel(Channel channel, HashMap itemDateMap)
{
  this.channel = channel;
  Iterator iter = channel.getItems().iterator();
  Date now = new Date();
  while(iter.hasNext())
    {
      Item item = (Item) iter.next();
      Integer hash = new Integer(item.calculateUniqueId());

      Date date = (Date) itemDateMap.get(hash);
      if(date == null)
        {
          log.warn("found item in channel but not in itemDateMap." + 
                   " set date to now. debug info follows:");
          log.debug("item: " + item);
          log.debug("item.hash: " + hash);
          
          date = now;
        }
      
      dateItemMap.put(date, item);
    }
}

public synchronized void updateChannel(Channel update)
{
  log.debug("begin updateChannel: " + update.getTitle());
  
  if(channel == null)
    {
      Date retrieved = new Date();
      setLastRetrievedDate(retrieved);

      setChannel(update);
      channelLevelChanges = true;
      
      return;
    }
  
  Date retrieved = new Date();
  log.debug("lastRetrievedDate: " + retrieved);
  setLastRetrievedDate(retrieved);

  List updateItems = update.getItems();
  List channelItemsSortedByHash = new ArrayList(channel.getItems());
  
  Comparator hashComparator = new HashComparator();
  Collections.sort(channelItemsSortedByHash, hashComparator);
  
  logChannelItems(channelItemsSortedByHash);
  logDateItemMap();
  
  List newItems = new ArrayList();
  Iterator iter = updateItems.iterator();
  while(iter.hasNext())
    {
      Item item = (Item) iter.next();
      if(item.hasTitle())
        log.debug("item: " + item.getTitle());
      log.debug("item.hash: " + item.calculateUniqueId());
      
      if(Collections.binarySearch(channelItemsSortedByHash,
                                  item,
                                  hashComparator) >= 0)
        {
          log.debug("already in list");
          continue;
        }
      
      log.debug("new item");
      
      dateItemMap.put(retrieved, item);

      //Add the new item to the list that is sorted by hash. That
      //should take care of the case that the same item is
      //(accidently) twice in update.
      channelItemsSortedByHash.add(item);
      
      //keep the list sorted in order to meet the contract of
      //Collections.binarySearch(...)
      Collections.sort(channelItemsSortedByHash, hashComparator);

      //Add it also to the list of new items
      newItems.add(item);
    }

  log.debug("newItems");
  Iterator newItemsIter = newItems.iterator();
  while(newItemsIter.hasNext())
    {
      Item item = (Item) newItemsIter.next();
      log.debug("item: " + item.getTitle());
    }

  List channelItems = channel.getItems();
  log.debug("channelItems");
  Iterator channelItemsIter = channelItems.iterator();
  while(channelItemsIter.hasNext())
    {
      Item item = (Item) channelItemsIter.next();
      log.debug("item: " + item.getTitle());
    }

  channelItems.addAll(0, newItems);

  log.debug("channelItems after");
  Iterator channelItemsIter2 = channelItems.iterator();
  while(channelItemsIter2.hasNext())
    {
      Item item = (Item) channelItemsIter2.next();
      log.debug("item: " + item.getTitle());
    }
  
  channelLevelChanges = !(channel.channelLevelEquals(update));
  
  //If changes occured at the channel level, we simply replace
  //our own channel object with a clone of update. Care needs to be
  //taken that we preserve the itemlist.
  if(channelLevelChanges)
    {
      try
        {
          channel = (Channel) update.clone();
          channel.setItems(channelItems);
        }
      catch(CloneNotSupportedException ex)
        {
          //can't occur, as clone() is part of the Channel interface
        }
    }

  log.debug("end updateChannel");
}

private void logChannelItems(List channelItems)
{
  log.debug("logChannelItems - begin");

  Iterator iter = channelItems.iterator();
  while(iter.hasNext())
    {
      Item item = (Item) iter.next();
      
      log.debug("item: " + item.getTitle());
      log.debug("item.hash: " + item.calculateUniqueId());
    }
  
  log.debug("logChannelItems - end");
}

private void logDateItemMap()
{
  log.debug("logDateItemMap - begin");
  
  Iterator dateItemIterator = dateItemMap.entrySet().iterator();
  
  while(dateItemIterator.hasNext())
    {
      Map.Entry entry = (Map.Entry) dateItemIterator.next();
      
      Date date = (Date) entry.getKey();
      log.debug("date: " + date);
      
      ArrayList itemList = (ArrayList) entry.getValue();
      
      Iterator itemIter = itemList.iterator();
      while(itemIter.hasNext())
        {
          Item item = (Item) itemIter.next();
      
          log.debug("item: " + item.getTitle());
          log.debug("item.hash: " + item.calculateUniqueId());
        }
    }

  log.debug("logDateItemMap - end");
}

public synchronized Date getLastRetrievedDate()
{
  return lastRetrievedDate;
}

public synchronized void setLastRetrievedDate(Date lastRetrievedDate)
{
  this.lastRetrievedDate = lastRetrievedDate;
}

public synchronized boolean hasChangedSince(Date since)
{
  return (since.compareTo(lastRetrievedDate) < 0);
}

public synchronized Collection getChangedDatesSince(Date since)
{
  SortedSet dateSet = new TreeSet(dateItemMap.keySet());
  Set sinceSet = dateSet.tailSet(since);

  return sinceSet;
}

public synchronized Collection getChangedDatesSince(Date beforeDate, 
                                                    Date afterDate)
{
  SortedSet dateSet = new TreeSet(dateItemMap.keySet());
  Set sinceSet = dateSet.subSet(beforeDate, afterDate);

  return sinceSet;
}

public synchronized boolean hasNewItems()
{
  return dateItemMap.containsKey(getLastRetrievedDate());
}

public synchronized List getNewItems()
{
  return getItems(getLastRetrievedDate());
}

public synchronized List getItems(Date since)
{
  List itemsSinceList = new ArrayList();
  Collection datesCol = getChangedDatesSince(since);
  
  return createItemListForDates(datesCol);
}

public synchronized List getItems(Date beforeDate, Date afterDate)
{
  Collection datesCol = getChangedDatesSince(beforeDate, afterDate);
  
  return createItemListForDates(datesCol);
}

private List createItemListForDates(Collection datesCol)
{
  List itemsForDatesList = new ArrayList();

  Iterator iter = datesCol.iterator();
  
  while(iter.hasNext())
    {
      ArrayList itemList = (ArrayList) dateItemMap.get(iter.next());
      itemsForDatesList.addAll(itemList);
    }
  
  return itemsForDatesList;
}

public synchronized Map getDateItemMap()
{
  return Collections.unmodifiableMap(dateItemMap);
}

public synchronized boolean hasChannelLevelChanges()
{
  return channelLevelChanges;
}

public synchronized void setChannelLevelChanges(boolean channelLevelChanges)
{
  this.channelLevelChanges = channelLevelChanges;
}

public synchronized void setLastModifiedString(String lastModifiedString)
{
  this.lastModifiedString = lastModifiedString;
}

public synchronized String getLastModifiedString()
{
  return lastModifiedString;
}

public synchronized boolean hasLastModifiedString()
{
  return (lastModifiedString != null);
}

public synchronized void setEtagString(String eTagString)
{
  this.eTagString = eTagString;
}

public synchronized String getEtagString()
{
  return eTagString;
}

public synchronized boolean hasEtagString()
{
  return (eTagString != null);
}

public String toString()
{
  return new ToStringBuilder(this)
    .append("channelUri", url).toString();
}

private class HashComparator
  implements Comparator
{
public int compare(Object o1,
                   Object o2)
{
  Item i1 = (Item) o1;
  Item i2 = (Item) o2;

  Integer hash1 = new Integer(i1.calculateUniqueId());
  Integer hash2 = new Integer(i2.calculateUniqueId());

  return hash1.compareTo(hash2);
}
}
}
