package org.jaggregator.rss.util;

import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import org.apache.commons.lang.exception.ExceptionUtils;
import org.apache.log4j.Category;
import org.jaggregator.rss.RssException;
import org.jaggregator.rss.Subscription;
import org.jaggregator.rss.SubscriptionExporter;
import org.jaggregator.rss.SubscriptionParser;
import org.jaggregator.rss.common.SubscriptionExporterImpl;
import org.jaggregator.rss.common.SubscriptionImpl;
import org.jaggregator.rss.common.SubscriptionParserImpl;
import org.quartz.JobDataMap;
import org.quartz.JobDetail;
import org.quartz.JobExecutionContext;
import org.quartz.JobExecutionException;
import org.quartz.Scheduler;
import org.quartz.SchedulerException;
import org.quartz.SchedulerFactory;
import org.quartz.SimpleTrigger;
import org.quartz.Trigger;
import org.quartz.helpers.TriggerUtils;
import org.quartz.impl.StdSchedulerFactory;

/**
 * This class maintains a simple repository of subscriptions. The
 * implementation is based on
 * <code>SubscriptionExporterImpl</code> and
 * <code>SubscriptionParserImpl</code>, which store and parse
 * <code>Subscriptions</code> using files and directories. All
 * subscriptions are stored under a specific directory the
 * <code>subscriptionsRoot</code>. Every subscription has its own
 * subdirectory below <code>subscriptionsRoot</code>. In fact,
 * this class assumes that every directory below
 * <code>subscriptionsRoot</code> contains a subscription that
 * can be parsed using a
 * <code>SubscriptionParserImpl</code>. Furthermore, the
 * directories below <code>subscriptionsRoot</code> have to
 * conform a strict naming convention: the name has to be the
 * urlencoded stringified URL of the subscription's channel. The
 * utility class<code>FileUtils</code> provides methods for
 * creating those names.
 *
 * @author <a href="mailto:jsz@jaggregator.org">Janek Schwarz</a>
 * @version $Revision: 1.21 $
 */
public class SubscriptionRegistry
{
private Category log = Category.getInstance(SubscriptionRegistry.class);

/**
 * Directory which serves as the root of all subscription
 * subdirectories.
 */
private File subscriptionsRoot;

/**
 * Map that maps URLs to subscriptions. The URL should equal
 * subscription.getUrl().
 */
private HashMap subscriptions;

/**
 * Default interval in milliseconds at which the retrieval of a
 * subscription's channel data will be scheduled.
 */
private long defaultRefreshRate = 3L * 60L * 60L * 1000L;

/**
 * Map, which maps the URL of a channel to its refresh
 * rate. This enables channel specific refresh rate.
 */
private HashMap channelRefreshRateMap;

/**
 * List of exceptions that occured during initialization of this
 * registry object. The list contains instances of
 * <code>DirectoryNameExceptionTupel</code>.
 */
private List initializationExceptionList;

/**
 * Indicates the initialization of the registry.
 */
private boolean initialized;

/**
 * Scheduler used for scheduling channel retrieval.
 */
Scheduler scheduler;

/**
 * List of registered event listeners.
 */
private List eventListeners = Collections.synchronizedList(new LinkedList());

/**
 * Indicates wether a proxy should be used.
 */
private boolean useProxy = false;

/**
 * Hostname of the proxy.
 */
private String proxyHost;

/**
 * Port of the proxy.
 */
private int proxyPort;

protected HashMap headerMap;
protected HashMap queryStringMap;

private EventDispatcherThread eventDispatcher = new EventDispatcherThread();

/**
 * Map that maps files to registry instances.
 */
private static HashMap registryMap;

static
{
  registryMap = new HashMap();
}

/**
 * Creates a new <code>SubscriptionRegistry</code> instance.
 *
 * @param subscriptionsRoot directory which servers as the root
 * of all subscription subdirectories
 */
SubscriptionRegistry(File subscriptionsRoot)
{
  this.subscriptionsRoot = subscriptionsRoot;
  this.channelRefreshRateMap = new HashMap();
  this.headerMap = new HashMap();
  this.queryStringMap = new HashMap();

  this.eventDispatcher.start();
}

/**
 * Get a <code>SubscriptionRegistry</code>. This method
 * guarantees at most one <code>SubscriptionRegistry</code> per
 * <code>subscriptionsRoot</code>.
 *
 * @param subscriptionsRoot directory which servers as the root
 * of all subscription subdirectories
 * @return a <code>SubscriptionRegistry</code>. This instance is
 * not initialized!
 */
public synchronized static SubscriptionRegistry getInstance(
  File subscriptionsRoot)
{
  SubscriptionRegistry registry;

  if(registryMap.containsKey(subscriptionsRoot))
    {
      registry = (SubscriptionRegistry) registryMap.get(subscriptionsRoot);
    }
  else
    {
      registry = new SubscriptionRegistry(subscriptionsRoot);
      registryMap.put(subscriptionsRoot, registry);
    }

  return registry;
}

public void addChannelRetrievalListener(ChannelRetrievalListener l)
{
  eventListeners.add(l);
}

public void removeChannelRetrievalListener(ChannelRetrievalListener l)
{
  eventListeners.remove(l);
}

/**
 * Set host and port of the proxy.
 *
 * @param proxyHost hostname
 * @param proxyPort port
 */
public void setProxy(String proxyHost, int proxyPort)
{
  this.useProxy = true;
  this.proxyHost = proxyHost;
  this.proxyPort = proxyPort;
}

public void addHttpHeader(String name, String value)
{
  headerMap.put(name, value);
}

public void addQueryStringPart(String name, String value)
{
  queryStringMap.put(name, value);
}

/**
 * Set the default refresh rate. If the interval is not an
 * even minute, it will be rounded to the next even minute.
 *
 * @param defaultRefreshRate interval
 */
public synchronized void setDefaultRefreshRate(long defaultRefreshRate)
{
  if(defaultRefreshRate < 0)
    throw new IllegalArgumentException("defaultRefreshRate < 0");

  if(defaultRefreshRate % 60 != 0)
    {
      Date date =
        TriggerUtils.getEvenMinuteDate(new Date(defaultRefreshRate));
      this.defaultRefreshRate = date.getTime();
    }
  else
    this.defaultRefreshRate = defaultRefreshRate;
}

/**
 * Set the default refresh rate.
 *
 * @return interval
 */
public synchronized long getDefaultRefreshRate()
{
  return defaultRefreshRate;
}

/**
 * Set the refresh rate for the retrieval of a channel
 * denoted by <code>channelUri</code>.  If the interval is not an
 * even minute, it will be rounded to the next even minute.
 */
public synchronized void setRefreshRate(URL channelUri,
                                        long refreshRate)
{
  Long refreshRateLong;

  if(defaultRefreshRate % 60 != 0)
    {
      Date date =
        TriggerUtils.getEvenMinuteDate(new Date(defaultRefreshRate));
      refreshRateLong = new Long(date.getTime());
    }
  else
    refreshRateLong = new Long(refreshRate);

  channelRefreshRateMap.put(channelUri, refreshRateLong);
}

/**
 * Get the refresh rate for the retrieval of a channel
 * denoted by <code>channelUri</code>. If no refresh rate was
 * defined for <code>channelUri</code> then the default refresh
 * rate is returned.
 *
 * @param channelUri the channel
 * @return the refresh rate
 */
public synchronized long getRefreshRate(URL channelUri)
{
  Long longObject = (Long) channelRefreshRateMap.get(channelUri);

  long ret = getDefaultRefreshRate();
  if(longObject != null)
    {
      ret = longObject.longValue();
    }

  return ret;
}

/**
 * If an error occured during initialization, the returned list
 * contains all exceptions that were thrown. The list is empty if
 * no errors happened. The list contains instances of
 * <code>DirectoryNameExceptionTupel</code>.
 *
 * @return a list of <code>DirectoryNameExceptionTupel</code>s
 */
public synchronized List getInitializationExceptionList()
{
  return initializationExceptionList;
}

/**
 * Parse the subscriptions. Retrieval of channel data is not
 * scheduled by this method. Use <code>scheduleRetrieval</code>
 * for that. In order to process the whole content below
 * <code>subscriptionsRoot</code> exceptions that occur during
 * initialization are kept in an internal list that can be
 * accessed via <code>getExceptionList()</code>.
 *
 * @return true if an error occured, false otherwise
 */
public synchronized boolean initialize()
{
  if(initialized)
    return false;

  String fileNames[] = subscriptionsRoot.list();
  this.subscriptions = new HashMap();
  this.initializationExceptionList = new LinkedList();

  for(int i = 0; i < fileNames.length; i++)
    {
      String directoryName = fileNames[i];

      try
        {
          URL channelUri = new URL(FileUtils.fileNameToUri(directoryName));
          SubscriptionParser parser =
            new SubscriptionParserImpl(subscriptionsRoot);
          Subscription subscription = parser.parse(channelUri);
          subscriptions.put(subscription.getUrl(), subscription);
        }
      catch(MalformedURLException ex)
        {
          initializationExceptionList.add(
            new DirectoryNameExceptionTupel(directoryName,
                                            ex));
        }
      catch(IOException ex)
        {
          initializationExceptionList.add(
            new DirectoryNameExceptionTupel(directoryName,
                                            ex));
        }
      catch(RssException ex)
        {
          initializationExceptionList.add(
            new DirectoryNameExceptionTupel(directoryName,
                                            ex));
        }
      catch(Throwable ex)
        {
          initializationExceptionList.add(
            new DirectoryNameExceptionTupel(directoryName,
                                            ex));
        }
    }

  initialized = true;

  return (initializationExceptionList.size() > 0);
}

private void initializeScheduler()
  throws RssException
{
  if(this.scheduler == null)
    {
      try
        {
          SchedulerFactory schedFact = new StdSchedulerFactory();
          this.scheduler = schedFact.getScheduler();
          this.scheduler.addGlobalJobListener(new JobListener());

          scheduler.start();
        }
      catch(SchedulerException ex)
        {
          throw new RssException(ex);
        }
    }
}

/**
 * Shuts a registry down. That includes removing the
 * subscriptions from the scheduler and -- optionaly -- saving
 * every subscription to disk. In order to shut down as cleanly
 * and complete as possible, occuring exceptions are kept in a
 * list and the shutdown continues if possible. The list of
 * exceptions is returned. Actually, the list contains instances
 * of <code>UrlExceptionTupel</code>s.
 *
 * @param registry <code>SubscriptionRegistry</code> which should shut down
 *
 * @param shouldSave if true, every subscription in the
 * <code>registry</code> will be saved to disk.
 *
 * @return a <code>List</code> of <code>UrlExceptionTupel</code>s
 * with exceptions that occured during shutdown. A size of 0
 * means that no exceptions occured.
 */
public synchronized static List shutdown(SubscriptionRegistry registry,
                                         boolean shouldSave)
{
  registryMap.remove(registry.subscriptionsRoot);

  return registry.shutdown(shouldSave);
}

/**
 * Does the actual shutdown. See shutdown(registry, shouldSave).
 */
private List shutdown(boolean shouldSave)
{
  List exceptionList = new LinkedList();

  try
    {
      if(scheduler != null)
        scheduler.pause();

      if(shouldSave)
        saveSubscriptions(exceptionList);

      if(scheduler != null)
        {
          String jobNames[] = scheduler.getJobNames(getGroupName());
          for(int i = 0; i < jobNames.length; i++)
            scheduler.deleteJob(jobNames[i], getGroupName());
        }

      if(scheduler != null)
        scheduler.start();
    }
  catch(SchedulerException ex)
    {
      exceptionList.add(new UrlExceptionTupel(null,
                                              ex));
    }
  catch(Throwable ex)
    {
      exceptionList.add(new UrlExceptionTupel(null,
                                              ex));
    }

  return exceptionList;
}

/**
 * Save all subscriptions registered with this
 * registry. Exceptions are not thrown when occured. Instead
 * their added to <code>exceptionList</code> and saving continues
 * if possible.
 *
 * @param exceptionList a <code>List</code> in which to store
 * exceptions.
 */
private void saveSubscriptions(List exceptionList)
{
  Iterator subscriptionIter = subscriptions.values().iterator();
  while(subscriptionIter.hasNext())
    {
      Subscription subscription = (Subscription) subscriptionIter.next();

      try
        {
          saveSubscription(subscription);
        }
      catch(RssException ex)
        {
          exceptionList.add(new UrlExceptionTupel(subscription.getUrl(),
                                                  ex));
        }
      catch(IOException ex)
        {
          exceptionList.add(new UrlExceptionTupel(subscription.getUrl(),
                                                  ex));
        }
      catch(Throwable ex)
        {
          exceptionList.add(new UrlExceptionTupel(subscription.getUrl(),
                                                  ex));
        }
    }
}

private void saveSubscription(Subscription subscription)
  throws RssException, IOException
{
  SubscriptionExporter exporter =
    new SubscriptionExporterImpl(subscriptionsRoot);

    exporter.export(subscription);
}

/**
 * Subscribe to a channel. Retrieval of data is scheduled
 * implicitely. The subscription data can optionally be saved to
 * the subscriptionsRoot.
 *
 * @param channelUri the <code>URL</code> to which to subscribe
 * @param shouldSave should the subscription data be saved?
 * @exception RssException if an error occurs during saving
 * @exception IOException if an error occurs during saving
 */
public synchronized void subscribe(URL channelUri, boolean shouldSave)
  throws RssException, IOException
{
  Subscription subscription = new SubscriptionImpl();
  subscription.setUrl(channelUri);
  subscription.setSubscriptionDate(new Date());
  subscriptions.put(channelUri, subscription);

  if(shouldSave)
    {
      SubscriptionExporter exporter =
        new SubscriptionExporterImpl(subscriptionsRoot);
      exporter.export(subscription);
    }

  if(scheduler != null)
    scheduleRetrieval(subscription);
}

/**
 * Unsubscribe from a channel. That includes removing the
 * subscription from the scheduler and deleting all related files
 * from disk.
 *
 * @param channelUri URI of the subscription to unsubscribe
 *
 * @return the subscription or null if no such subscription
 * existed
 *
 * @exception IOException if an error occured when deleting the
 * file from disk.
 */
public synchronized Subscription unsubscribe(URL channelUri)
  throws IOException
{
  Subscription subscription =
    (Subscription) subscriptions.remove(channelUri);

  if(subscription != null)
    {
      try
        {
          if(scheduler != null)
            scheduler.deleteJob(createJobName(subscription),
                                getGroupName());
        }
      catch(SchedulerException ex)
        {
          //ignore it
          log.error(ExceptionUtils.getStackTrace(ex));
        }

      FileUtils.deleteDirectory(
        new File(subscriptionsRoot,
                 FileUtils.uriToFileName(channelUri.toString())));
    }

  return subscription;
}

/**
 * Unsubscribe from all channels who's uri is contained in
 * channelUriCollection. That includes removing the
 * subscription from the scheduler and deleting all related files
 * from disk.
 *
 * @param channelUriCollection a <code>Collection</code> of URL
 * objects.
 *
 * @exception IOException if an error occured when deleting the
 * file from disk.
 */
public synchronized void unsubscribe(Collection channelUriCollection)
  throws IOException
{
  Iterator iter = channelUriCollection.iterator();

  while(iter.hasNext())
    unsubscribe((URL) iter.next());
}

/**
 * Test whether a subscription exists.
 *
 * @param channelUri URL of the subscribed channel
 * @return true if it exists, false otherwise
 */
public synchronized boolean containsSubscription(URL channelUri)
{
  return (getSubscription(channelUri) != null);
}

/**
 * Get a subscription
 *
 * @param channelUri URL of the subscribed channel

 * @return the subscription or null if no such subscription
 * exists.
 */
public synchronized Subscription getSubscription(final URL channelUri)
{
  return (Subscription) subscriptions.get(channelUri);
}

/**
 * Get an unmodifiable list of subscriptions.
 *
 * @return an unmodifiable <code>Collection</code> containing all
 * subscriptions. Modifying the resulting exception yields an
 * <code>UnsupportedOperationException</code>.
 */
public synchronized Collection getSubscriptions()
{
  return Collections.unmodifiableCollection(subscriptions.values());
}

/**
 * Get the URIs of all subscribed channels.
 *
 * @return <code>Set</code> containing <code>URL</code>s.
 */
public synchronized Set getSubscriptionUris()
{
  return subscriptions.keySet();
}

/**
 * Get an unmodifiable Map that maps URIs to Subscriptions.
 *
 * @return a <code>Map</code> value
 */
public synchronized Map getSubscriptionsMap()
{
  return Collections.unmodifiableMap(subscriptions);
}

/**
 * Return the next time when a channel will be updated.
 *
 * @return a <code>Date</code> value
 */
public Date getNextUpdateTime()
  throws RssException
{
  try
    {
      String groupName = getGroupName();
      String[] triggerNames = scheduler.getTriggerNames(groupName);

      Date nextUpdateDate = null;
      for(int i = 0; i < triggerNames.length; i++)
        {
          Trigger trigger = scheduler.getTrigger(triggerNames[i],
                                                 groupName);

          Date date = trigger.getNextFireTime();

          if(nextUpdateDate == null)
            nextUpdateDate = date;
          else if(nextUpdateDate.compareTo(date) > 0)
            nextUpdateDate = date;
        }

      return nextUpdateDate;
    }
  catch(SchedulerException ex)
    {
      throw new RssException(ex);
    }
}

/**
 * Schedule retrieval of new data for each subscription
 * registered with this instance. This method should only called
 * once. Retrieval stops when <code>shutdown</code> is
 * called. The method makes sure that retrieval of channel data
 * is scheduled only at even minutes.
 *
 * @exception RssException if an error occurs
 */
public synchronized void scheduleRetrieval()
  throws RssException
{
  initializeScheduler();

  Iterator iter = subscriptions.values().iterator();

  while(iter.hasNext())
    {
      Subscription s = (Subscription) iter.next();
      scheduleRetrieval(s);
    }
}

/**
 * Schedule retrieval of new channel data for
 * <code>subscription</code>.
 *
 * @param subscription <code>Subscription</code> to be scheduled
 * @exception RssException if an error occurs
 */
private void scheduleRetrieval(Subscription subscription)
  throws RssException
{
  try
    {
      long refreshRate = getRefreshRate(subscription.getUrl());

      JobDetail jobDetail = createJobDetail(subscription);
      Date startTime = calculateFirstStartTime(subscription,
                                               refreshRate);
      Trigger trigger = createTrigger(subscription,
                                      startTime,
                                      refreshRate);

      scheduler.scheduleJob(jobDetail, trigger);

      log.debug(subscription.getUrl() + " scheduled for: " + startTime);
    }
  catch(SchedulerException ex)
    {
      throw new RssException(ex);
    }
}

/**
 * Creates a <code>JobDetail</code> instance for
 * <code>subscription</code>.
 *
 * @param subscription a <code>Subscription</code> value
 * @return a <code>JobDetail</code> value
 */
JobDetail createJobDetail(Subscription subscription)
{
  JobDetail jobDetail = new JobDetail(createJobName(subscription),
                                      getGroupName(),
                                      ChannelRetrievalJob.class);

  JobDataMap dataMap = jobDetail.getJobDataMap();

  dataMap.put("subscriptionRegistry", this);
  dataMap.put("subscription", subscription);
  dataMap.put("useProxy", useProxy);

  if(useProxy)
    {
      dataMap.put("proxyHost", proxyHost);
      dataMap.put("proxyPort", proxyPort);
    }

  dataMap.put("headerMap", headerMap);
  dataMap.put("queryStringMap", queryStringMap);

  return jobDetail;
}

/**
 * Given a <code>Subscription</code> object and its interval for
 * repeating the retrieval of data, this method calls the first
 * time at which retrieval is to be scheduled.
 *
 * @param subscription a <code>Subscription</code> value
 *
 * @param refreshRate time in milliseconds that has to pass
 * before another retrieval of data is started.
 *
 * @return start time
 */
Date calculateFirstStartTime(Subscription subscription,
                             long refreshRate)
{
  Date now = TriggerUtils.getEvenMinuteDate(null);
  Date lastRetrievedDate = subscription.getLastRetrievedDate();
  Date startTime = now;
  if(lastRetrievedDate != null)
    {
      if(lastRetrievedDate.getTime() + refreshRate < now.getTime())
        startTime = now;
      else
        {
          Date date = TriggerUtils.getEvenMinuteDate(lastRetrievedDate);
          startTime = new Date(date.getTime() + refreshRate);
        }
    }

  return startTime;
}

/**
 * Creates a trigger, which triggers retrieval of channel data.
 *
 * @param subscription a <code>Subscription</code> value
 * @param startTime time of first retrieval of data
 * @param refreshRate time in milliseconds that has to pass
 * before another retrieval of data is started.
 * @return a <code>Trigger</code>
 */
Trigger createTrigger(Subscription subscription,
                      Date startTime,
                      long refreshRate)
{
  return new SimpleTrigger(createTriggerName(subscription),
                           getGroupName(),
                           startTime,
                           null,
                           SimpleTrigger.REPEAT_INDEFINITELY,
                           refreshRate);
}

/**
 * Create a unique name for a retrieval job.
 *
 * @param s <code>Subscription</code> to be named
 * @return a unique name for use as a job name
 */
String createJobName(Subscription s)
{
  return "Job: " + s.getUrl().toString();
}

/**
 * Create a unique name for a trigger.
 *
 * @param s <code>Subscription</code> to be named
 * @return a unique name for use as a trigger name
 */
String createTriggerName(Subscription s)
{
  return "Trigger: " + s.getUrl().toString();
}

/**
 * Get the name this registry uses as a group name when creating
 * jobs and triggers.
 *
 * @return a <code>String</code> value
 */
String getGroupName()
{
  return subscriptionsRoot.getAbsolutePath();
}

private class JobListener
  implements org.quartz.JobListener
{
public String getName()
{
  return JobListener.class.toString();
}

public void jobToBeExecuted(JobExecutionContext context)
{
  JobDetail jobDetail = context.getJobDetail();

  if(jobDetail.getGroup().equals(getGroupName()))
    {
      JobDataMap dataMap = jobDetail.getJobDataMap();
      Subscription subscription = (Subscription) dataMap.get("subscription");

      ChannelRetrievalEvent event =
        new ChannelRetrievalEvent(this,
                                  subscription.getUrl(),
                                  ChannelRetrievalEvent.STARTED);

      eventDispatcher.dispatchEvent(event);

      log.debug("going to retrieve: " + subscription.getUrl());
    }
}

/**
 * Check wether execution of a ChannelRetrievalJob was
 * successful. If not schedule a retry.
 */
public void jobWasExecuted(JobExecutionContext context,
                           JobExecutionException jobException)
{
  JobDetail jobDetail = context.getJobDetail();

  if(jobDetail.getGroup().equals(getGroupName()))
    {
      JobDataMap dataMap = context.getJobDetail().getJobDataMap();
      Subscription subscription = (Subscription) dataMap.get("subscription");

      ChannelRetrievalEvent.Progress progress;

      if(jobException == null)
        {
          log.debug(subscription.getUrl() + " retrieved.");
          progress = ChannelRetrievalEvent.FINISHED_SUCCESS;

          SubscriptionRegistry registry =
            (SubscriptionRegistry) dataMap.get("subscriptionRegistry");

          try
            {
              registry.saveSubscription(subscription);
              log.debug(subscription.getUrl() + " saved.");
            }
          catch(IOException ex)
            {
              log.error("caught exception: " + ex.getMessage());
              log.debug("subscription: " + subscription.getUrl());
              log.debug(ExceptionUtils.getStackTrace(ex));
            }
          catch(RssException ex)
            {
              log.error("caught exception: " + ex.getMessage());
              log.debug("subscription: " + subscription.getUrl());
              log.debug(ExceptionUtils.getStackTrace(ex));
            }
          catch(Throwable ex)
            {
              log.error("caught exception: " + ex.getMessage());
              log.debug("subscription: " + subscription.getUrl());
              log.debug(ExceptionUtils.getStackTrace(ex));
            }
        }
      else
        {
          //RetrieveExceptions encapsule IO- and HttpExceptions. For
          //those, it makes sense to retry retrieval in a couple of
          //minutes.
          if(jobException.getUnderlyingException()
             instanceof RetrieveException)
            {
              //TODO(JS): schedule a retry

              log.error(ExceptionUtils.getStackTrace(
                          jobException.getUnderlyingException()));
            }

          progress = ChannelRetrievalEvent.FINISHED_FAILURE;
        }

      ChannelRetrievalEvent event =
        new ChannelRetrievalEvent(this,
                                  subscription.getUrl(),
                                  progress);

      eventDispatcher.dispatchEvent(event);
    }
}
}

public class UrlExceptionTupel
{
private URL url;
private Throwable exception;

UrlExceptionTupel(URL url,
                  Throwable exception)
{
  this.url = url;
  this.exception = exception;
}

/**
 * Return the URL. Might be null.
 *
 * @return an <code>URL</code> value
 */
public URL getUrl()
{
  return url;
}

/**
 * Return the exception.
 *
 * @return a <code>Throwable</code> value
 */
public Throwable getException()
{
  return exception;
}
}


public class DirectoryNameExceptionTupel
{
private String directoryName;
private Throwable exception;

DirectoryNameExceptionTupel(String directoryName,
                            Throwable exception)
{
  this.directoryName = directoryName;
  this.exception = exception;
}

/**
 * Return the directory name. Might be null.
 *
 * @return an <code>String</code> value
 */
public String getDirectoryName()
{
  return directoryName;
}

/**
 * Return the exception.
 *
 * @return a <code>Throwable</code> value
 */
public Throwable getException()
{
  return exception;
}
}

private class EventDispatcherThread
  extends Thread
{
private List eventList = new LinkedList();

public synchronized void run()
{
  try
    {
      while(true)
        {
          while(eventList.size() == 0)
            wait();

          Iterator eventIter = eventList.iterator();
          while(eventIter.hasNext())
            {
              ChannelRetrievalEvent event =
                (ChannelRetrievalEvent) eventIter.next();

              Iterator listenerIter = eventListeners.iterator();
              while(listenerIter.hasNext())
                {
                  ChannelRetrievalListener listener =
                    (ChannelRetrievalListener) listenerIter.next();

                  if(event.getProgress().equals(ChannelRetrievalEvent.STARTED))
                    listener.channelRetrievalStarted(event);
                  else
                    listener.channelRetrievalFinished(event);
                }

              eventIter.remove();
            }
        }
    }
  catch(InterruptedException ex)
    {
      Thread.currentThread().interrupt();
    }
}

public synchronized void dispatchEvent(ChannelRetrievalEvent e)
{
  eventList.add(e);
  notifyAll();
}
}
}

